from hexbytes import HexBytes

from eth_permissions.access_control import Component, Role, get_registry


def test_simple_role_hash():
    role = Role("LEVEL1_ROLE")
    assert role.hash == HexBytes("0xbf372ca3ebecfe59ac256f17697941bbe63302aced610e8b0e3646f743c7beb2")


def test_unknown_role():
    role = Role.from_hash(HexBytes("0x55435dd261a4b9b3364963f7738a7a662ad9c84396d64be3365284bb7f0a5041"))
    assert role.name == "UNKNOWN ROLE: 0x5543...5041"


def test_component():
    component = Component(HexBytes("0x8c5f6aEB655D687929a82c5d430Ec56abaDdc0c8"), "Test")
    assert str(component) == "Test"


def test_component_unknown_name():
    component = Component(HexBytes("0x8c5f6aEB655D687929a82c5d430Ec56abaDdc0c8"))
    assert str(component) == "Component<0x8c5f...c0c8>"


def test_component_role():
    component = Component(HexBytes("0x8c5f6aEB655D687929a82c5d430Ec56abaDdc0c8"), "Test")
    role = Role("POLICY_CREATOR_ROLE", component=component)
    assert role.hash == HexBytes("0x1ca1414ebf49990bc4f22652af37d8d535c4dc522f3557d79ede5d6b1c1c9ffb")
    assert str(role) == "Role:POLICY_CREATOR_ROLE@Test"


def test_registry():
    kr_names = ["GUARDIAN_ROLE", "LEVEL1_ROLE", "LEVEL2_ROLE", "LEVEL3_ROLE"]
    registry = get_registry()
    registry.add_roles([Role(name) for name in kr_names])

    assert (
        registry.get("0x0000000000000000000000000000000000000000000000000000000000000000")
        == Role.default_admin()
    )
    assert registry.get("0x55435dd261a4b9b3364963f7738a7a662ad9c84396d64be3365284bb7f0a5041") == Role(
        "GUARDIAN_ROLE"
    )
    assert registry.get("0x5ec196419322369c6bac572d883f4d990ae5ec82f7d93cdf89b85dbb05b63c27") == Role(
        "LEVEL3_ROLE"
    )
    assert registry.get("0xa82e22387fca439f316d78ca566f383218ab8ae1b3e830178c9c82cbd16749c0") == Role(
        "LEVEL2_ROLE"
    )
    assert registry.get("0xbf372ca3ebecfe59ac256f17697941bbe63302aced610e8b0e3646f743c7beb2") == Role(
        "LEVEL1_ROLE"
    )


def test_registry_with_components():
    kr_names = ["RESOLVER_ROLE", "PRICER_ROLE"]
    components = [Component(HexBytes("0x8c5f6aEB655D687929a82c5d430Ec56abaDdc0c8"), "TestRM")]

    registry = get_registry()
    registry.add_roles([Role(name) for name in kr_names])
    registry.add_components(components)

    assert (
        registry.get("0x0000000000000000000000000000000000000000000000000000000000000000")
        == Role.default_admin()
    )
    assert registry.get("0x92a19c77d2ea87c7f81d50c74403cb2f401780f3ad919571121efe2bdb427eb1") == Role(
        "RESOLVER_ROLE"
    )
    assert registry.get("0xc6823861ee2bb2198ce6b1fd6faf4c8f44f745bc804aca4a762f67e0d507fd8a") == Role(
        "PRICER_ROLE"
    )
    assert registry.get("0x1efef69cb7b7efbed1b57c9a070d0e45faca403bad919571121efe2bdb427eb1") == Role(
        "RESOLVER_ROLE", component=components[0]
    )
    assert registry.get("0x4add528a8b76da60a54e9da02ca189e5fe2a8574804aca4a762f67e0d507fd8a") == Role(
        "PRICER_ROLE", component=components[0]
    )



from itertools import zip_longest

import functions_framework
from environs import Env
from hexbytes import HexBytes

from eth_permissions.access_control import Component, Role, get_registry
from eth_permissions.graph import build_graph

env = Env()
env.read_env()

KNOWN_ROLES = env.list("KNOWN_ROLES", ["GUARDIAN_ROLE", "LEVEL1_ROLE", "LEVEL2_ROLE", "LEVEL3_ROLE"])
KNOWN_COMPONENTS = env.list("KNOWN_COMPONENTS", [])
KNOWN_COMPONENT_NAMES = env.list("KNOWN_COMPONENT_NAMES", [])  # TODO: we should get the names from chain

CORS_HEADERS = {
    "Access-Control-Allow-Origin": "*",
}

if len(KNOWN_COMPONENTS) < len(KNOWN_COMPONENT_NAMES):
    raise RuntimeError("Can't have a component name without address")

get_registry().add_roles([Role(name) for name in KNOWN_ROLES])
get_registry().add_components(
    [
        Component(HexBytes(address), name)
        for address, name in zip_longest(KNOWN_COMPONENTS, KNOWN_COMPONENT_NAMES)
    ]
)


@functions_framework.http
def permissions_graph(request):
    try:
        address = request.args["address"]
    except KeyError:
        return {"error": "address is required"}, 400

    graph = build_graph(address)
    return (graph.source, 200, CORS_HEADERS)



import itertools
from collections import defaultdict
from dataclasses import dataclass
from datetime import timedelta
from typing import Dict, List, Literal, Optional, Set

from eth_typing import ChecksumAddress, HexStr
from eth_utils import add_0x_prefix, to_checksum_address

MAX_UINT64 = 2**64 - 1


@dataclass(frozen=True)
class Role:
    id: int
    label: str = None
    grant_delay: timedelta = timedelta(0)

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        if not isinstance(other, Role):
            return False
        return self.id == other.id

    def as_dict(self):
        return {
            "id": self.id,
            "label": self.label,
            "grant_delay": self.grant_delay,
        }

    def from_dict(data: dict) -> "Role":
        return Role(
            id=data["id"],
            label=data["label"],
            grant_delay=(
                timedelta(seconds=data["grant_delay"])
                if not isinstance(data["grant_delay"], timedelta)
                else data["grant_delay"]
            ),
        )


@dataclass(frozen=True)
class Target:
    address: ChecksumAddress
    closed: bool = False
    admin_delay: timedelta = timedelta(0)

    def __hash__(self):
        return hash(self.address)

    def __eq__(self, other):
        if not isinstance(other, Target):
            return False
        return self.address == other.address

    def as_dict(self):
        return {
            "address": self.address,
            "closed": self.closed,
            "admin_delay": self.admin_delay,
        }

    def from_dict(data: dict) -> "Target":
        return Target(
            address=data["address"],
            closed=data["closed"],
            admin_delay=timedelta(seconds=data["admin_delay"]),
        )


@dataclass(frozen=True)
class RoleMember:
    address: ChecksumAddress
    execution_delay: timedelta = timedelta(0)

    def __hash__(self):
        return hash((self.address))

    def __eq__(self, other):
        if not isinstance(other, RoleMember):
            return False
        return self.address == other.address

    def as_dict(self):
        return {
            "address": self.address,
            "execution_delay": self.execution_delay,
        }


@dataclass(frozen=True)
class SelectorRole:
    role: Role
    selector: HexStr

    def __hash__(self):
        return hash((self.role, self.selector))

    def __eq__(self, other):
        if not isinstance(other, SelectorRole):
            return False
        return self.role == other.role and self.selector == other.selector


class AccessManager:
    roles: Dict[int, Role]  # id -> Role
    targets: Dict[ChecksumAddress, Target]  # address -> Target

    role_members: Dict[int, Set[RoleMember]]  # role_id -> Members
    role_admins: Dict[int, Role]  # role_id -> Admin
    role_guardians: Dict[int, Role]  # role_id -> Guardian

    target_allowed_roles: Dict[ChecksumAddress, Set[SelectorRole]]  # target -> Selector Allowed Role

    ADMIN_ROLE = Role(label="ADMIN_ROLE", id=0)
    PUBLIC_ROLE = Role(label="PUBLIC_ROLE", id=MAX_UINT64)

    def __init__(self):
        self.roles = {
            self.ADMIN_ROLE.id: self.ADMIN_ROLE,
            self.PUBLIC_ROLE.id: self.PUBLIC_ROLE,
        }
        self.targets = {}
        self.role_members = defaultdict(set)
        self.role_admins = {}
        self.role_guardians = {}
        self.target_allowed_roles = defaultdict(set)

    @classmethod
    def from_events(cls, events: List[dict]) -> "AccessManager":
        """Loads a current state access manager as defined by the events."""
        am = cls()
        for role_id, events in itertools.groupby(events, key=lambda e: e["args"].roleId):
            for event in events:
                if event["event"] == "RoleGranted":
                    # RoleGranted(uint64 indexed roleId, address indexed account, uint32 delay, uint48 since, bool newMember);  # noqa
                    am.grant_role(
                        Role(role_id),
                        to_checksum_address(event["args"].account),
                        timedelta(seconds=event["args"].delay),
                    )
                elif event["event"] == "RoleRevoked":
                    # RoleRevoked(uint64 indexed roleId, address indexed account)
                    am.revoke_role(Role(role_id), to_checksum_address(event["args"].account))
                elif event["event"] == "RoleGuardianChanged":
                    # RoleGuardianChanged(uint64 indexed roleId, uint64 indexed guardian)
                    am.set_role_guardian(Role(role_id), Role(event["args"].guardian))
                elif event["event"] == "RoleAdminChanged":
                    # RoleAdminChanged(uint64 indexed roleId, uint64 indexed admin)
                    am.set_role_admin(Role(role_id), Role(event["args"].admin))
                elif event["event"] == "RoleLabel":
                    # RoleLabel(uint64 indexed roleId, string label)
                    am.label_role(Role(role_id), event["args"].label)
                elif event["event"] == "TargetFunctionRoleUpdated":
                    # TargetFunctionRoleUpdated(address indexed target, bytes4 selector, uint64 indexed roleId)  # noqa
                    am.set_target_function_role(
                        am.get_target(to_checksum_address(event["args"].target)),
                        {add_0x_prefix(HexStr(event["args"].selector.hex()))},
                        Role(role_id),
                        fix=True,
                    )
                elif event["event"] == "RoleGrantDelayChanged":
                    # RoleGrantDelayChanged(uint64 indexed roleId, uint32 delay, uint48 since);
                    am.set_grant_delay(Role(role_id), timedelta(seconds=event["args"].delay))
                elif event["event"] == "TargetClosed":
                    # TargetClosed(address indexed target, bool closed)
                    am.set_target_closed(
                        am.get_target(to_checksum_address(event["args"].target)), event["args"].closed
                    )
                elif event["event"] == "TargetAdminDelayUpdated":
                    # TargetAdminDelayUpdated(address indexed target, uint32 delay, uint48 since)
                    am.set_target_admin_delay(
                        am.get_target(to_checksum_address(event["args"].target)),
                        timedelta(seconds=event["args"].delay),
                    )
                else:
                    raise RuntimeError(f"Unexpected event {event.name} for role {role_id}")
        return am

    def get_role(self, role_id: int) -> Optional[Role]:
        return self.roles.get(role_id)

    def get_role_members(self, role: Role) -> Set[RoleMember]:
        return self.role_members.get(role.id, set())

    def get_role_admin(self, role: Role) -> Optional[Role]:
        return self.role_admins.get(role.id, self.ADMIN_ROLE)

    def get_role_guardian(self, role: Role) -> Optional[Role]:
        return self.role_guardians.get(role.id, self.ADMIN_ROLE)

    def get_target(self, address: ChecksumAddress) -> Optional[Target]:
        return self.targets.get(address, Target(address, False, timedelta(0)))

    def get_target_allowed_role(self, address: ChecksumAddress, selector: HexStr) -> Optional[Role]:
        for selector_role in self.target_allowed_roles.get(address, set()):
            if selector_role.selector == selector:
                return selector_role.role

    def get_all_target_selectors(self, target: Target) -> Set[HexStr]:
        return {
            selector_role.selector for selector_role in self.target_allowed_roles.get(target.address, set())
        }

    def get_all_role_targets(self, role: Role) -> Dict[Target, Set[HexStr]]:
        ret = {}
        for target_address, selector_roles in self.target_allowed_roles.items():
            selectors = {
                selector_role.selector for selector_role in selector_roles if selector_role.role == role
            }
            if selectors:
                ret[self.get_target(target_address)] = selectors
        return ret

    def label_role(self, role: Role, label: str):
        if role.id in self.roles:
            if self.roles[role.id].label != label:
                self.roles[role.id] = Role(role.id, label, self.roles[role.id].grant_delay)
        else:
            self.roles[role.id] = Role(role.id, label, role.grant_delay)

        return self.roles[role.id]

    def set_grant_delay(self, role: Role, delay: timedelta):
        if role.id in self.roles:
            if self.roles[role.id].grant_delay != delay:
                self.roles[role.id] = Role(role.id, self.roles[role.id].label, delay)
        else:
            self.roles[role.id] = Role(role.id, role.label, delay)

        return self.roles[role.id]

    def grant_role(self, role: Role, member: ChecksumAddress, execution_delay: timedelta = timedelta(0)):
        if role.id not in self.roles:
            self.roles[role.id] = role

        new_member = RoleMember(member, execution_delay)

        if new_member in self.role_members.get(role.id, set()):
            # Member already has the role, remove it in case the delay has changed
            self.role_members[role.id].remove(new_member)
        self.role_members[role.id].add(new_member)

        return new_member

    def revoke_role(self, role: Role, member: ChecksumAddress):
        if role.id not in self.roles:
            self.roles[role.id] = role

        member = RoleMember(member)
        if member in self.role_members.get(role.id, set()):
            self.role_members[role.id].remove(member)

    def set_role_admin(self, role: Role, admin: Role):
        if role.id not in self.roles:
            self.roles[role.id] = role
        if admin.id not in self.roles:
            self.roles[admin.id] = admin

        role = self.roles[role.id]
        admin = self.roles[admin.id]

        self.role_admins[role.id] = admin

    def set_role_guardian(self, role: Role, guardian: Role):
        if role.id not in self.roles:
            self.roles[role.id] = role
        if guardian.id not in self.roles:
            self.roles[guardian.id] = guardian

        role = self.roles[role.id]
        guardian = self.roles[guardian.id]

        self.role_guardians[role.id] = guardian

    def set_target_function_role(self, target: Target, selectors: Set[HexStr], role: Role, fix: bool = False):
        if role.id not in self.roles:
            self.roles[role.id] = role

        role = self.roles[role.id]

        if not selectors:
            return
        for selector in selectors:
            selector_role = SelectorRole(role, selector)
            if selector_role not in self.target_allowed_roles.get(target.address, set()):
                self.target_allowed_roles[target.address].add(selector_role)
                # Finds other roles that have the same target/selector
                other_roles = [
                    sr
                    for sr in self.target_allowed_roles[target.address]
                    if sr.selector == selector and sr.role != role
                ]
                if not other_roles:
                    continue
                if fix:
                    # When reading on-chain operations, just remove the old roles (override)
                    for sr in other_roles:
                        self.target_allowed_roles[target.address].remove(sr)
                else:
                    # When reading from the manifest, report the error, so we can't have the same method
                    # delegated to two roles
                    other_roles = [x.role for x in other_roles]
                    raise RuntimeError(
                        f"{target}->{selector} assigned to role {role} and also to other_roles: {other_roles}"
                    )

    def set_target_closed(self, target: Target, closed: bool):
        if target.address not in self.targets:
            self.targets[target.address] = Target(target.address, closed, target.admin_delay)
        else:
            self.targets[target.address] = Target(
                target.address, closed, self.targets[target.address].admin_delay
            )

        return self.targets[target.address]

    def set_target_admin_delay(self, target: Target, delay: timedelta):
        if target.address not in self.targets:
            self.targets[target.address] = Target(target.address, target.closed, delay)
        else:
            self.targets[target.address] = Target(target.address, self.targets[target.address].closed, delay)

        return self.targets[target.address]

    def as_dict(self) -> dict:
        """Returns the current access manager state as a dict with only base types.

        The same dict can be used to create an AccessManager instance with the `from_dict` method.
        """
        return {
            "roles": {
                role.id: {
                    **role.as_dict(),
                    "guardian": self.get_role_guardian(role).as_dict(),
                    "admin": self.get_role_admin(role).as_dict(),
                    "members": [member.as_dict() for member in self.get_role_members(role)],
                    "targets": {
                        target.address: selectors
                        for target, selectors in self.get_all_role_targets(role).items()
                    },
                }
                for role in self.roles.values()
            },
            "targets": {address: target.as_dict for address, target in self.targets.items()},
        }

    @classmethod
    def from_dict(cls, data: dict) -> "AccessManager":
        am = cls()
        for address, target_data in data.get("targets", {}).items():
            address = to_checksum_address(address)
            am.targets[address] = Target.from_dict(target_data)

        for role_id, role_data in data.get("roles", {}).items():
            role_id = int(role_id)
            am.label_role(Role(role_id), role_data["label"])
            am.set_grant_delay(Role(role_id), timedelta(seconds=role_data["grant_delay"]))
            am.set_role_admin(Role(role_id), Role(**role_data["admin"]))
            am.set_role_guardian(Role(role_id), Role(**role_data["guardian"]))
            for member_data in role_data["members"]:
                am.grant_role(
                    Role(role_id), member_data["address"], timedelta(seconds=member_data["execution_delay"])
                )

            for target_address, selectors in role_data["targets"].items():
                target = am.get_target(to_checksum_address(target_address))
                am.set_target_function_role(target, selectors, Role(role_id))
        return am


@dataclass
class Operation:
    # labelRole(uint64 roleId, string calldata label)
    # grantRole(uint64 roleId, address account, uint32 executionDelay)
    # revokeRole(uint64 roleId, address account)
    # setRoleAdmin(uint64 roleId, uint64 admin)
    # setRoleGuardian(uint64 roleId, uint64 guardian)
    # setGrantDelay(uint64 roleId, uint32 newDelay)
    # setTargetFunctionRole(address target, bytes4[] calldata selectors, uint64 roleId)
    # setTargetAdminDelay(address target, uint32 newDelay)
    op: Literal[
        "grantRole",
        "revokeRole",
        "setRoleGuardian",
        "setRoleAdmin",
        "labelRole",
        "setTargetFunctionRole",
        "setGrantDelay",
        "setTargetAdminDelay",
    ]
    args: dict

    def as_dict(self):
        return {"op": self.op, "args": self.args}



import itertools
from collections import defaultdict
from datetime import timedelta
from warnings import warn

from ethproto.wrappers import ETHWrapper, get_provider

from . import abis
from . import access_manager as am
from .access_control import get_registry


class BaseEventStream:
    ABI = None

    def __init__(self, contract_address, provider=None):
        self.contract_address = contract_address
        self._event_stream = None

        if provider is None:
            provider = get_provider("w3")
        self.provider = provider

    def _get_contract_wrapper(self):
        contract = self.provider.w3.eth.contract(address=self.contract_address, abi=self.ABI)
        return ETHWrapper.connect(contract)

    def _get_events(self, event_names):
        contract_wrapper = self._get_contract_wrapper()
        return self.provider.get_events(contract_wrapper, event_names)

    @property
    def stream(self):
        if self._event_stream is None:
            self._load_stream()
        return self._event_stream


class AccessControlEventStream(BaseEventStream):
    ABI = abis.OZ_ACCESS_CONTROL

    def _load_stream(self):
        events = self._get_events(["RoleGranted", "RoleRevoked"])  # TODO: RoleAdminChanged

        event_stream = []
        for event in events:
            event_stream.append(
                {
                    "role": get_registry().get("0x" + event.args.role.hex()),
                    "subject": event.args.account,
                    "requester": event.args.sender,
                    "order": (event.blockNumber, event.logIndex),
                    "event": event.event,
                }
            )
        self._event_stream = sorted(event_stream, key=lambda e: (e["role"].hash, e["order"]))

    @property
    def snapshot(self):
        snapshot = defaultdict(set)
        for role, events in itertools.groupby(self.stream, key=lambda e: e["role"].hash):
            for event in events:
                if event["event"] == "RoleGranted":
                    snapshot[role].add(event["subject"])
                elif event["event"] == "RoleRevoked":
                    try:
                        snapshot[role].remove(event["subject"])
                        if not snapshot[role]:
                            snapshot.pop(role)
                    except KeyError:
                        warn(f"WARNING: can't remove ungranted role {role} from {event['subject']}")
                else:
                    raise RuntimeError(f"Unexpected event {event.name} for role {role}")
        return [
            {"role": get_registry().get(role), "members": list(members)} for role, members in snapshot.items()
        ]


class AccessManagerEventStream(BaseEventStream):
    ABI = abis.OZ_ACCESS_MANAGER

    def _load_stream(self):
        events = self._get_events(
            [
                "RoleGranted",
                "RoleRevoked",
                "RoleGuardianChanged",
                "RoleAdminChanged",
                "RoleLabel",
                "TargetFunctionRoleUpdated",
                "RoleGrantDelayChanged",
                "TargetClosed",
                "TargetAdminDelayUpdated",
            ]
        )

        event_stream = [
            {
                "event": e.event,
                "args": e.args,
                "order": (e.blockNumber, e.logIndex),
            }
            for e in events
        ]

        self._event_stream = sorted(event_stream, key=lambda e: e["order"])

    @property
    def snapshot(self) -> am.AccessManager:
        """Returns a snapshot of the current permissions setup.

        The snapshot is an instance of AccessManager with the current state of the permissions.
        """
        return am.AccessManager.from_events(self.stream)

    @property
    def snapshot_dict(self) -> dict:
        return self.snapshot.as_dict()

    def compare(self, snapshot: am.AccessManager):
        """Compares the current snapshot with the given one. Returns the differences.

        snapshot must be a snapshot previously obtained from the snapshot property.

        The differences are returned as a series of operations to apply to bring the current state to the
        snapshot state.
        """
        current = self.snapshot
        differences = []

        for role_id, current_role in current.roles.items():
            snapshot_role = snapshot.roles.get(role_id, am.Role(id=role_id, label=""))

            if current_role.label != snapshot_role.label:
                differences.append(
                    am.Operation("labelRole", {"roleId": current_role, "label": snapshot_role.label})
                )
            if current.get_role_admin(current_role) != snapshot.get_role_admin(snapshot_role):
                differences.append(
                    am.Operation(
                        "setRoleAdmin",
                        {"roleId": current_role, "admin": snapshot.get_role_admin(snapshot_role)},
                    )
                )
            if current.get_role_guardian(current_role) != snapshot.get_role_guardian(snapshot_role):
                differences.append(
                    am.Operation(
                        "setRoleGuardian",
                        {"roleId": current_role, "guardian": snapshot.get_role_guardian(snapshot_role)},
                    )
                )
            if current_role.grant_delay != snapshot_role.grant_delay:
                differences.append(
                    am.Operation(
                        "setGrantDelay",
                        {"roleId": current_role, "newDelay": snapshot_role.grant_delay},
                    )
                )

            if current.get_role_members(current_role) != snapshot.get_role_members(snapshot_role):
                for member in current.get_role_members(current_role) - snapshot.get_role_members(
                    snapshot_role
                ):
                    differences.append(
                        am.Operation("revokeRole", {"roleId": current_role, "account": member.address})
                    )

                for member in snapshot.get_role_members(snapshot_role) - current.get_role_members(
                    current_role
                ):
                    differences.append(
                        am.Operation(
                            "grantRole",
                            {
                                "roleId": current_role,
                                "account": member.address,
                                "executionDelay": member.execution_delay,
                            },
                        )
                    )

            # For common members just need to check the execution delay is properly set
            for common_member in current.get_role_members(current_role) & snapshot.get_role_members(
                snapshot_role
            ):
                # Can't get te members from iterating the intersection, because we need to distinguish them
                current_member = [
                    member for member in current.get_role_members(current_role) if member == common_member
                ][0]
                snapshot_member = [
                    member for member in snapshot.get_role_members(snapshot_role) if member == common_member
                ][0]
                if current_member.execution_delay != snapshot_member.execution_delay:
                    differences.append(
                        am.Operation(
                            "grantRole",
                            {
                                "roleId": current_role,
                                "account": snapshot_member.address,
                                "executionDelay": snapshot_member.execution_delay,
                            },
                        )
                    )

        snapshot_targets = set(snapshot.targets.values())
        current_targets = set(current.targets.values())
        if current_targets != snapshot_targets:
            # The targets not present in the snapshot need to be reset to default values
            for target in current_targets - snapshot_targets:
                if target.closed:
                    differences.append(am.Operation("setTargetClosed", {"target": target, "closed": False}))
                if target.admin_delay != timedelta(0):
                    differences.append(am.Operation("setTargetAdminDelay", {"target": target, "newDelay": 0}))

            # The targets in snapshot but missing from current need to be properly configured
            for target in snapshot_targets - current_targets:
                if target.closed:
                    differences.append(
                        am.Operation("setTargetClosed", {"target": target, "closed": target.closed})
                    )
                if target.admin_delay != timedelta(0):
                    differences.append(
                        am.Operation(
                            "setTargetAdminDelay", {"target": target, "newDelay": target.admin_delay}
                        )
                    )

            # The common targets need to be checked for matching config
            for common_target in current_targets & snapshot_targets:
                current_target = [target for target in current_targets if target == common_target][0]
                snapshot_target = [target for target in snapshot_targets if target == common_target][0]
                if current_target.closed != snapshot_target.closed:
                    differences.append(
                        am.Operation(
                            "setTargetClosed", {"target": current_target, "closed": snapshot_target.closed}
                        )
                    )
                if current_target.admin_delay != snapshot_target.admin_delay:
                    differences.append(
                        am.Operation(
                            "setTargetAdminDelay",
                            {"target": current_target, "newDelay": snapshot_target.admin_delay},
                        )
                    )

        current_selectors = {
            (target, selector_role.selector)
            for target, selector_roles in current.target_allowed_roles.items()
            for selector_role in selector_roles
        }
        snapshot_selectors = {
            (target, selector_role.selector)
            for target, selector_roles in snapshot.target_allowed_roles.items()
            for selector_role in selector_roles
        }

        if current_selectors != snapshot_selectors:
            # The target -> selectors not present in the snapshot need to be assigned to ADMIN_ROLE
            for target, selector in current_selectors - snapshot_selectors:
                differences.append(
                    am.Operation(
                        "setTargetFunctionRole",
                        {
                            "target": target,
                            "selectors": {selector},
                            "roleId": snapshot.ADMIN_ROLE,
                        },
                    )
                )

            # The target -> selectors in snapshot missing from current need to be set on current
            for target, selector in snapshot_selectors - current_selectors:
                role = snapshot.get_target_allowed_role(target, selector)
                if role == snapshot.ADMIN_ROLE:
                    continue
                differences.append(
                    am.Operation(
                        "setTargetFunctionRole",
                        {
                            "target": target,
                            "selectors": {selector},
                            "roleId": role,
                        },
                    )
                )

        # The common target -> selectors need to be checked for matching role
        for common_target, common_selector in current_selectors & snapshot_selectors:
            current_role = current.get_target_allowed_role(common_target, common_selector)
            snapshot_role = snapshot.get_target_allowed_role(common_target, common_selector)
            if current_role != snapshot_role:
                differences.append(
                    am.Operation(
                        "setTargetFunctionRole",
                        {
                            "target": common_target,
                            "selectors": {common_selector},
                            "roleId": snapshot_role,
                        },
                    )
                )

        return differences



import argparse
import json
from itertools import zip_longest

from environs import Env
from hexbytes import HexBytes

from eth_permissions import access_manager as am
from eth_permissions.access_control import Component, Role, get_registry
from eth_permissions.chaindata import AccessManagerEventStream
from eth_permissions.graph import build_graph
from eth_permissions.utils import safe_serializer

env = Env()
env.read_env()

KNOWN_ROLES = env.list("KNOWN_ROLES", ["GUARDIAN_ROLE", "LEVEL1_ROLE", "LEVEL2_ROLE", "LEVEL3_ROLE"])
KNOWN_COMPONENTS = env.list("KNOWN_COMPONENTS", [])
KNOWN_COMPONENT_NAMES = env.list("KNOWN_COMPONENT_NAMES", [])


parser = argparse.ArgumentParser(
    prog="eth-permissions", description="Command line tool for auditing smart contract permissions"
)
parser.add_argument(
    "-t",
    "--type",
    help="Contract type - AccessManager (default) or AccessControl",
    default="AccessManager",
    required=False,
    choices=["AccessManager", "AccessControl"],
)
parser.add_argument(
    "-c",
    "--compare-snapshot",
    help=(
        "Compare the current snapshot with the one in the given file. "
        "Prints out the differences in json format."
    ),
)
parser.add_argument("-o", "--output", help="Output file. Only valid for graph output")
parser.add_argument(
    "-f",
    "--format",
    required=False,
    default=None,
    help=(
        "Output format. Only valid for graph output. "
        "If ommitted it will be determined from the output file name extension"
    ),
)
parser.add_argument(
    "-v",
    "--view",
    action="store_true",
    required=False,
    help=(
        "If specified, the rendered file image will be opened with the default viewer. "
        "Only valid for graph output."
    ),
)
parser.add_argument("address", help="The contract's address")


def load_registry():
    get_registry().add_roles([Role(name) for name in KNOWN_ROLES])
    get_registry().add_components(
        [
            Component(HexBytes(address), name)
            for address, name in zip_longest(KNOWN_COMPONENTS, KNOWN_COMPONENT_NAMES)
        ]
    )


def main():
    args = parser.parse_args()

    if args.type == "AccessManager":
        event_stream = AccessManagerEventStream(args.address)
        # print(
        #     "\n".join(
        #         f"{e['event']} | " + " ".join(f"{k}={v}" for k, v in e["args"].items())
        #         for e in event_stream.stream
        #     )
        # )

        if args.compare_snapshot:
            with open(args.compare_snapshot, "r") as f:
                reference_snapshot = json.load(f)
            snapshot = am.AccessManager.from_dict(reference_snapshot)
            comparison = event_stream.compare(snapshot)
            print(json.dumps(comparison, indent=2, default=safe_serializer))
        else:
            snapshot = event_stream.snapshot_dict
            print(json.dumps(snapshot, indent=2, default=safe_serializer))
        return

    if not args.output:
        raise ValueError("Output file must be specified")

    load_registry()

    graph = build_graph(args.address)

    kwargs = {}
    if args.format:
        kwargs["format"] = args.format

    graph.render(outfile=args.output, cleanup=True, view=args.view, **kwargs)



OZ_ACCESS_CONTROL = [
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "bytes32", "name": "role", "type": "bytes32"},
            {"indexed": True, "internalType": "address", "name": "account", "type": "address"},
            {"indexed": True, "internalType": "address", "name": "sender", "type": "address"},
        ],
        "name": "RoleGranted",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "bytes32", "name": "role", "type": "bytes32"},
            {"indexed": True, "internalType": "address", "name": "account", "type": "address"},
            {"indexed": True, "internalType": "address", "name": "sender", "type": "address"},
        ],
        "name": "RoleRevoked",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "bytes32", "name": "role", "type": "bytes32"},
            {"indexed": True, "internalType": "bytes32", "name": "previousAdminRole", "type": "bytes32"},
            {"indexed": True, "internalType": "bytes32", "name": "newAdminRole", "type": "bytes32"},
        ],
        "name": "RoleAdminChanged",
        "type": "event",
    },
]

OZ_ACCESS_MANAGER = [
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "uint64", "name": "roleId", "type": "uint64"},
            {"indexed": True, "internalType": "address", "name": "account", "type": "address"},
            {"indexed": False, "internalType": "uint32", "name": "delay", "type": "uint32"},
            {"indexed": False, "internalType": "uint48", "name": "since", "type": "uint48"},
            {"indexed": False, "internalType": "bool", "name": "newMember", "type": "bool"},
        ],
        "name": "RoleGranted",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "uint64", "name": "roleId", "type": "uint64"},
            {"indexed": True, "internalType": "address", "name": "account", "type": "address"},
        ],
        "name": "RoleRevoked",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "uint64", "name": "roleId", "type": "uint64"},
            {"indexed": True, "internalType": "uint64", "name": "guardian", "type": "uint64"},
        ],
        "name": "RoleGuardianChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "uint64", "name": "roleId", "type": "uint64"},
            {"indexed": True, "internalType": "uint64", "name": "admin", "type": "uint64"},
        ],
        "name": "RoleAdminChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "uint64", "name": "roleId", "type": "uint64"},
            {"indexed": False, "internalType": "string", "name": "label", "type": "string"},
        ],
        "name": "RoleLabel",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "address", "name": "target", "type": "address"},
            {"indexed": False, "internalType": "bytes4", "name": "selector", "type": "bytes4"},
            {"indexed": True, "internalType": "uint64", "name": "roleId", "type": "uint64"},
        ],
        "name": "TargetFunctionRoleUpdated",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "uint64", "name": "roleId", "type": "uint64"},
            {"indexed": False, "internalType": "uint32", "name": "delay", "type": "uint32"},
            {"indexed": False, "internalType": "uint48", "name": "since", "type": "uint48"},
        ],
        "name": "RoleGrantDelayChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "address", "name": "target", "type": "address"},
            {"indexed": False, "internalType": "bool", "name": "closed", "type": "bool"},
        ],
        "name": "TargetClosed",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "address", "name": "target", "type": "address"},
            {"indexed": False, "internalType": "uint32", "name": "delay", "type": "uint32"},
            {"indexed": False, "internalType": "uint48", "name": "since", "type": "uint48"},
        ],
        "name": "TargetAdminDelayUpdated",
        "type": "event",
    },
]



from dataclasses import dataclass
from itertools import zip_longest

from eth_utils import add_0x_prefix
from hexbytes import HexBytes
from web3 import Web3

from .utils import ellipsize


@dataclass
class Component:
    address: HexBytes
    name: str = None

    def __str__(self):
        if self.name:
            return self.name
        return f"Component<{ellipsize(add_0x_prefix(self.address.hex()))}>"

    def to_json(self):
        return {"address": add_0x_prefix(self.address.hex()), "name": self.name}


class Role:
    def __init__(self, name, component: Component = None):
        self.name = name
        self.component = component
        self._role_hash = HexBytes(Web3.keccak(text=name))

    @classmethod
    def from_hash(cls, hash: HexBytes, component: Component = None):
        ret = cls(name=f"UNKNOWN ROLE: {ellipsize(add_0x_prefix(hash.hex()))}", component=component)
        ret._role_hash = hash
        return ret

    @classmethod
    def component_role_from_role(cls, hash: HexBytes, role):
        role_xor_addr = hash.hex()[:-24]
        base_hash = role.hash.hex()[:-24]
        component_address = int(role_xor_addr, 16) ^ int(base_hash, 16)
        component_address = HexBytes(hex(component_address))

        ret = cls(name=role.name, component=Component(address=component_address))
        ret._role_hash = role.hash
        return ret

    @classmethod
    def default_admin(cls):
        ret = cls("DEFAULT_ADMIN_ROLE")
        ret._role_hash = HexBytes("0x" + "0" * 64)
        return ret

    @property
    def hash(self) -> HexBytes:
        if not getattr(self, "_hash", None):
            if self.component is None:
                self._hash = self._role_hash
            else:
                self._hash = HexBytes(
                    bytes(
                        [
                            cbyte ^ rbyte
                            for cbyte, rbyte in zip_longest(
                                self.component.address, self._role_hash, fillvalue=0
                            )
                        ]
                    )
                )
        return self._hash

    def to_json(self) -> dict:
        return {
            "name": self.name,
            "hash": add_0x_prefix(self.hash.hex()),
            "component": self.component.to_json() if self.component else None,
        }

    def __str__(self):
        ret = f"Role:{self.name}"
        if self.component:
            ret += f"@{self.component}"
        return ret

    def __repr__(self):
        ret = f"Role('{self.name}')"
        if self.component:
            ret += f"@{self.component}"
        return ret

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, Role):
            return False
        return self.hash == __o.hash

    def __hash__(self):
        return hash(self.hash)


class Registry:
    def __init__(self):
        self._map = {}
        self.add(Role.default_admin())

    def add_roles(self, roles):
        for role in roles:
            self.add(role)

    def add_components(self, components):
        for component in components:
            for role in list(self._map.values()):
                self.add(Role(role.name, component=component))

    def add(self, role):
        self._map[role.hash] = role
        if role.component and role._role_hash not in self._map:
            base_role = Role(role.name)
            base_role._role_hash = role._role_hash
            self._map[role._role_hash] = base_role

    def get(self, hash):
        if isinstance(hash, str):
            hash = HexBytes(hash)
        if hash in self._map:
            return self._map[hash]
        # Try to match the last part of the hash
        hash_tail = hash.hex()[-24:]
        base_role = next(
            (r for h, r in self._map.items() if h.hex()[-24:] == hash_tail and r.component is None), None
        )
        if base_role:
            # It's a component role of an unknown component
            return Role.component_role_from_role(hash, base_role)
        else:
            return Role.from_hash(hash)


_registry = None


def get_registry():
    global _registry
    if _registry is None:
        _registry = Registry()
    return _registry



import re
from collections import defaultdict
from datetime import timedelta


def ellipsize(hexa):
    """Ellipsize a hex string

    >>> ellipsize("0x4d68Cf31d222270b18E406AFd6A42719a62a0785") -> "0x4d68...0785"
    """
    match = re.match(r"(0x[0-9a-fA-F]{4})([0-9a-fA-F]*?)([0-9a-fA-F]{4})$", hexa)
    if not match:
        raise ValueError(f"Expected a hex string beggining with 0x, got {hexa}")
    return match.expand(r"\1...\3")


class ExplorerAddress:
    EXPLORER_URL_TEMPLATE = "https://polygonscan.com/address/{address}"

    @classmethod
    def get(cls, address):
        return cls.EXPLORER_URL_TEMPLATE.format(address=address)


def safe_serializer(obj):
    if hasattr(obj, "as_dict"):
        return obj.as_dict()
    if isinstance(obj, set):
        return list(obj)
    if isinstance(obj, defaultdict):
        return {k: list(v) for k, v in obj.items()}
    if isinstance(obj, timedelta):
        return int(obj.total_seconds())
    raise TypeError(f"Object of type {type(obj)} is not serializable")



import graphviz

from .chaindata import AccessControlEventStream
from .utils import ExplorerAddress, ellipsize


def build_graph(contract_address):
    stream = AccessControlEventStream(contract_address)

    dot = graphviz.Digraph("Permissions")
    dot.attr(rankdir="RL", splines="ortho")
    dot.attr("node", style="rounded", shape="box")

    dot.node(
        "CONTRACT",
        URL=ExplorerAddress.get(contract_address),
        target="_blank",
        style="filled",
        fillcolor="green",
        shape="hexagon",
        fontcolor="blue",
    )

    for item in stream.snapshot:
        graphviz.quoting.quote(str(item["role"]))
        dot.node(item["role"].hash.hex(), str(item["role"]), tooltip=item["role"].hash.hex())
        # dot.edge(item["role"].hash.hex(), "CONTRACT")

        for member in item["members"]:
            dot.node(
                member,
                ellipsize(member),
                tooltip=member,
                URL=ExplorerAddress.get(member),
                target="_blank",
                style="filled",
                shape="hexagon",
                fontcolor="blue",
            )
            dot.edge(member, item["role"].hash.hex())

    return dot
